#include "atlas.h"

#include "map.h"
#include "build.h"
#include "glbdef.h"
#include "attrib.h"
#include "rooms/allrooms.h"
#include <stdio.h>

#define WORLDWIDTH 10
#define WORLDHEIGHT 10


extern MOB *glbAvatar;

static ATLAS *theAtlas;

ATLAS *
glbAtlas()
{
    if (!theAtlas)
	theAtlas = new ATLAS();
    return theAtlas;
}

ATLAS::ATLAS()
{
    myClearSquare = SQUARE_WALL;
    myPhase = 0;
}

ATLAS::~ATLAS()
{
    for (int i = 0; i < myMaps.entries(); i++)
	delete myMaps(i);

    for (int i = 0; i < myRoomDefs.entries(); i++)
	delete myRoomDefs(i);
}

void
ATLAS::createNewWorld()
{
    if (theAtlas)
    {
	delete theAtlas;
	theAtlas = 0;
    }
    ATLAS	*world = glbAtlas();

    for (int i = 0; i < NUM_ALLROOMDEFS; i++)
    {
	MAP	*room = new MAP(*glb_allroomdefs[i]);
	world->registerMap(room);
    }

    {
	ROOM_DEF	*def = BUILD::build(BUILD_WORLD, 0, "world", "world");
	MAP		*room = new MAP(*def);
	world->registerMap(room);
	world->registerDef((ROOM_DEF_NC *)def);
    }

    glbAvatar = new MOB(MOB_AVATAR);
    glbAvatar->setFacing(DIR::south());

#if 0
    MAP *map = world->findMap("home");
    POS p = map->findRandomItem(ITEM_BED);
    glbAvatar->moveTo(p);
#else
    MAP *map = world->findMap("world");
    POS p = map->findRandomPassable();
    glbAvatar->moveTo(p);
#endif
}

void
ATLAS::resetAll()
{
    for (int i = 0; i < myMaps.entries(); i++)
	myMaps(i)->resetAll();
}

void
ATLAS::registerMap(MAP *map)
{
    int	idx = myMaps.append(map);

    map->setAtlas(this, idx);
}

void
ATLAS::registerDef(ROOM_DEF_NC *def)
{
    myRoomDefs.append(def);
}

int
ATLAS::getDaystars() const
{
    const MAP	*world = findMap("world");
    if (!world)
	return 0;
    return world->level();
}

void
ATLAS::addDaystar()
{
    MAP			*world = findMap("world");
    if (!world)
	return;
    world->incLevel();

    if (world->level() == 4)
    {
	// Add the portal!
	int	ntries = 50;
	while (1)
	{
	    int		x = rand_choice(world->width());
	    POS		p = world->buildPos(x, 10);
	    if (p.item())
	    {
		ntries--;
		if (ntries < 0)
		{
		    if (p.item()->getDefinition() == ITEM_ALTAR)
		    {
			continue;
		    }
		}
		else
		    continue;
	    }
	    ITEM		*portal = new ITEM(ITEM_PORTAL_BOTTOM);
	    portal->moveTo(p);
	    p = p.north();
	    portal = new ITEM(ITEM_PORTAL_TOP);
	    portal->moveTo(p);
	    break;
	}
    }
}

const MAP *
ATLAS::findMap(const char *name) const
{
    if (!name)
	return 0;
    for (int i = 0; i < getNumMaps(); i++)
    {
	if (!strcmp(name, map(i)->name()))
	    return map(i);
    }
    return 0;
}

const MAP *
ATLAS::findMapByQuest(const char *name) const
{
    if (!name)
	return 0;
    for (int i = 0; i < getNumMaps(); i++)
    {
	if (!map(i)->defn().quest)
	    continue;
	if (!strcmp(name, map(i)->defn().quest))
	    return map(i);
    }
    return 0;
}

MAP *
ATLAS::findMap(const char *name)
{
    if (!name)
	return 0;
    for (int i = 0; i < getNumMaps(); i++)
    {
	if (!strcmp(name, map(i)->name()))
	    return map(i);
    }
    return 0;
}

MAP *
ATLAS::findMapByQuest(const char *name)
{
    if (!name)
	return 0;
    for (int i = 0; i < getNumMaps(); i++)
    {
	if (!map(i)->defn().quest)
	    continue;
	if (!strcmp(name, map(i)->defn().quest))
	    return map(i);
    }
    return 0;
}

void		 
ATLAS::resetFlag(FLAG_NAMES flag, bool val)
{
    for (int i = 0; i < myMaps.entries(); i++)
    {
	myMaps(i)->resetFlag(flag, val);
    }
}

void		 
ATLAS::buildFOV(POS center, int rad)
{
    resetFlag(FLAG_FOV, false);
    resetFlag(FLAG_SHADOW, true);

#if 1
    center.setFlag(FLAG_FOV, true);
    center.setFlag(FLAG_SHADOW, false);

    DIR		dir;
    FORALL_DIR(dir)
    {
	WAVEFRONT	front(center, dir);
	PTRLIST<int>	vis, newvis;

	vis.append(3);

	int		frontcenter = 1;
	for (int r = 1; r < rad; r++)
	{
	    newvis.clear();
	    for (int i = 0; i < front.entries(); i++)
	    {
		int	islit = 0;

		if (i < frontcenter)
		{
		    islit |= vis(i);
		    if (i)
			islit |= vis(i-1);
		}
		else if (i == frontcenter)
		{
		    islit = vis(i-1);
		}
		else
		{
		    islit |= vis(i-2);
		    if (i <= vis.entries())
			islit |= vis(i-1);
		}
		if (islit && front(i).valid())
		    front(i).setFlag(FLAG_FOV, true);
		if ((islit == 3) && front(i).valid())
		    front(i).setFlag(FLAG_SHADOW, false);

		if (!front(i).defn().transparent)
		{
		    if (front(i).defn().semitrans)
		    {
			islit >>= 1;
		    }
		    else
			islit = 0;
		}
		newvis.append(islit);
	    }
	    // Expand to cover walls
	    for (int i = 0; i < front.entries(); i++)
	    {
		int	islit = 0;
#if 1
		if (i)
		    islit |= newvis(i-1);
#endif
#if 1
		if (i < front.entries()-1)
		    islit |= newvis(i+1);
#endif
		if (islit && front(i).valid())
		    front(i).setFlag(FLAG_FOV, true);
		if ((islit == 3) && front(i).valid())
		    front(i).setFlag(FLAG_SHADOW, false);
	    }
	    vis = newvis;
	    front.advance();
	    frontcenter++;
	}
    }
#endif
}

void
ATLAS::clear(SQUARE_NAMES square)
{
    for (ATLAS::iterator it = begin(); !it.done(); it.advance())
    {
	it.pos().setTile(square);
    }
}

ATLAS::iterator::iterator(const ATLAS *atlas)
{
    myAtlas = atlas;
    mapid = -1;
    myDone = false;
    advance();
}

void
ATLAS::iterator::advance()
{
    if (myDone)
	return;

    if (!it.done())
	it.advance();
    if (it.done())
    {
	mapid++;
	if (mapid >= myAtlas->getNumMaps())
	{
	    myDone = true;
	    return;
	}
	it = myAtlas->map(mapid)->begin(); 
    }
}

POS
ATLAS::iterator::pos() const
{
    return it.pos();
}

bool
ATLAS::iterator::done() const
{
    return myDone;
}

ATLAS::iterator
ATLAS::begin() const
{
    ATLAS::iterator	it(this);

    return it;
}

void
ATLAS::heartbeat()
{
    myPhase++;
    for (int id = 0; id < getNumMaps(); id++)
    {
	MAP	*m = map(id);
	if (m)
	    m->heartbeat(myPhase);
    }
}

void
ATLAS::clearMessages()
{
    for (int id = 0; id < getNumMaps(); id++)
    {
	MAP	*m = map(id);
	if (m)
	    m->clearMessages();
    }
}
