#pragma once

#include "glbdef.h"
#include "rand.h"
#include "ptrlist.h"

class MAP;
class MOB;
class ITEM;

class POS;
typedef PTRLIST<POS> POSLIST;
typedef PTRLIST<ITEM *> ITEMLIST;

class DIR
{
public:
    DIR() {}
    DIR(int angle) { myAngle = angle; }
    static DIR		north() { return DIR(0); }
    static DIR		west() { return DIR(1); }
    static DIR		south() { return DIR(2); }
    static DIR		east() { return DIR(3); }
    static DIR		rand() { return DIR(rand_choice(4)); }

    DIR			counterclock() const { int angle = myAngle + 1; if (angle == 4) angle = 0; return DIR(angle); }
    DIR			clock() const { int angle = myAngle - 1; if (angle == -1) angle = 3; return DIR(angle); }

    DIR			inverse() const { int angle = myAngle + 2; if (angle >= 4) angle -= 4; return DIR(angle); }

    int			angle() const { return myAngle; }

    bool		operator==(const DIR &b) const
    {
	return myAngle == b.myAngle;
    }
    bool		operator!=(const DIR &b) const
    {
	return myAngle != b.myAngle;
    }

protected:
    int		myAngle;
};

inline DIR operator-(DIR a, DIR b)
{
    return DIR((a.angle() - b.angle()) & 3);
}

inline DIR operator+(DIR a, DIR b)
{
    return DIR((a.angle() + b.angle()) & 3);
}

#define FORALL_DIR(dir) \
    for (int lcl_angle = 0; dir = DIR(lcl_angle), lcl_angle < 4; lcl_angle++)

typedef PTRLIST<DIR> DIRLIST;

class POS
{
public:
    SQUARE_NAMES	tile() const;
    const SQUARE_DEF	&defn() const;
    MOB			*mob() const;
    ITEM 		*item() const;
    ITEM_NAMES		 itemname() const;
    int 		 itemlist(ITEMLIST &list) const;
    bool		 valid() const;

    void 		setTile(SQUARE_NAMES square);
    void		setFlag(FLAG_NAMES flag, bool val);
    bool		getFlag(FLAG_NAMES flag) const;

    bool		passable() const { return defn().passable; }
    bool		surface() const;

    void		convertToLava();
    void		setDarkness();
    void		removeDarkness();

    void		findLinkedSquares(POSLIST &list);
    bool		toggleLinkedSquares();
    bool		toggleSquare();

    // For when links means something other than links.
    void		getLinks(PTRLIST<int> &links) const;

    POS			delta(DIR dir) const;

    POS			operator+=(DIR dir)
			{ *this = delta(dir); return *this; }
    POS			operator-=(DIR dir)
			{ *this = delta(dir.inverse()); return *this; }

    POS			up() const;
    POS			down() const;
    POS			north() const { return delta(DIR(0)); }
    POS			west() const { return delta(DIR(1)); }
    POS			south() const { return delta(DIR(2)); }
    POS			east() const { return delta(DIR(3)); }

    int			level() const;

    static POS		 invalidPos() 
    {
	POS		p;
	p.myMap = 0;
	p.myX = -1;
	p.myY = -1;
	return p;
    }

    bool		operator==(const POS &b) const
    {
	return myX == b.myX && myY == b.myY && myMap == b.myMap;
    }
    bool		operator!=(const POS &b) const
    {
	return ! ((*this) == b);
    }

    // Never use these.
    MAP		*map() const { return myMap; }
    int		 x() const { return myX; }
    int		 y() const { return myY; }
protected:
    int		myX, myY;
    MAP		*myMap;

    friend class MAP;
};


class PATH
{
public:
    PATH() {}
    PATH(DIR dir) { myPath.append(dir); }

    POS apply(POS p) const;
    PATH		operator+=(DIR dir)
			{ myPath.append(dir); return *this; }
    PATH		operator+=(const PATH &path)
			{ myPath.append(path.myPath); return *this; }
protected:
    DIRLIST	myPath;
};

inline POS operator+(POS p, DIR d)
{
    return p.delta(d);
}

inline POS operator+(POS p, const PATH &path)
{
    return path.apply(p);
}


inline POS operator-(POS p, DIR d)
{
    return p.delta(d.inverse());
}

class WAVEFRONT
{
public:
    WAVEFRONT() {}
    WAVEFRONT(POS center, DIR dir, bool inclusive = false);

    POS 	operator()(int idx) const { return myFront(idx); }
    int		entries() const { return myFront.entries(); }
    void	advance();
protected:
    POSLIST	myFront;
    DIR		myDir;
    bool	myInclusive;
};

class CIRCLEFRONT
{
public:
    CIRCLEFRONT(POS center);

    POS 	operator()(int idx) const;
    int		entries() const;
    void	advance();
protected:
    WAVEFRONT	myFronts[4];
    POS		myCenter;
    bool	myAtCenter;
};

